/* Eventually I have to copy this to all the files when I officially release the source. */

/*
    PolyPaint - Multiuser Painting Program
    Copyright (C) 2004 smart_idiot

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <allegro.h>
#include <stdlib.h>
#include <stdio.h> /* For (v)snprintf */
#include <math.h>

#ifdef USE_ALFONT
#include <alfont.h>
ALFONT_FONT *alfont = 0;
#endif

#include "mem.h"
#include "assert.h"
#include "command.h"
#include "commandqueue.h"
#include "network.h"
#include "draw.h"
#include "rng.h"
#include "tools.h"
#include "config.h"
#include "version.h"
#include "user.h"
#include "chat.h"

BITMAP *preview = 0;
BITMAP *local_image = 0, *remote_image = 0, *scratch_buffer = 0;
BITMAP *buffer = 0, *cursor = 0;

DirtyRectangles dirt;

int inmenu = 0;
Connection connection;
int can_connect = 0, connected = 0;
CommandQueue pending;
int drew_users = 0;

uint8_t userid = 0;

Colour colour = {0.0, 0.0, 0.0};
HSV hsv = {0.0, 0.0, 0.0};

int shadowColour(Colour c)
 {
  float v = c.r;
  
  if(v < c.g)
   v = c.g;
  
  if(v < c.b)
   v = c.b;
  
  if(v > 120.0)
   return makecol32(0, 0, 0);
  else
   return makecol32(255, 255, 255);
 }

void redrawCursor(void)
 {
  int bg = shadowColour(colour), fg = makecol32(colour.r, colour.g, colour.b);
  
  show_mouse(0);
  clear_to_color(cursor, MASK_COLOR_32);
  
  hline(cursor, 1, 3, 5, bg);
  vline(cursor, 3, 1, 5, bg);
  
  hline(cursor, 0, 2, 4, fg);
  vline(cursor, 2, 0, 4, fg);
  
  set_mouse_sprite(cursor);
  set_mouse_sprite_focus(2, 2);
  
  show_mouse(screen);
 }

#define PATH_MAX_POINTS 256
int16_t y[PATH_MAX_POINTS];
int16_t x[PATH_MAX_POINTS];
uint32_t points = 0;

float angleDiff(float x1, float y1, float x2, float y2)
 {
  float d = sqrt(x1*x1+y1*y1);
  x1 /= d;
  y1 /= d;
  d = sqrt(x2*x2+y2*y2);
  x2 /= d;
  y2 /= d;
  return sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2));
 }

BITMAP *menu;
void redrawMenuIcons(void);
void redrawMenu(void);

Colour colour_history_start[32] = {{255, 255, 255},  /* 0  - white */
                                   {160, 160, 160},  /* 1  - light grey */
                                   {128, 128, 128},  /* 2  - grey */
                                   {64,  64,  64 },  /* 3  - dark grey */
                                   {0,   0,   0  },  /* 4  - black */
                                   {255, 0,   0  },  /* 5  - red */
                                   {255, 128, 0  },  /* 6  - orange */
                                   {255, 255, 0  },  /* 7  - yellow */
                                   {128, 255, 0  },  /* 8  - lime */
                                   {0,   255, 0  },  /* 9  - green */
                                   {0,   255, 128},  /* 10 - green/teal/something */
                                   {0,   255, 255},  /* 11 - teal */
                                   {0,   128, 255},  /* 12 - blue/teal/something */
                                   {0,   0,   255},  /* 13 - blue */
                                   {128, 0,   255},  /* 14 - blue-magenta */
                                   {255, 0,   255},  /* 15 - magenta */
                                   {255, 0,   128},  /* 16 - red-magenta */
                                   {128, 0,   0  },  /* 17 - red */
                                   {128, 64,  0  },  /* 18 - orange */
                                   {128, 128, 0  },  /* 19 - yellow */
                                   {64,  128, 0  },  /* 20 - lime */
                                   {0,   128, 0  },  /* 21 - green */
                                   {0,   128, 64 },  /* 22 - green/teal/something */
                                   {0,   128, 128},  /* 23 - teal */
                                   {0,   64,  128},  /* 24 - blue/teal/something */
                                   {0,   0,   128},  /* 25 - blue */
                                   {64,  0,   128},  /* 26 - blue-magenta */
                                   {128, 0,   128},  /* 27 - magenta */
                                   {128, 0,   64},   /* 28 - red-magenta */
                                   {255, 128, 128},  /* 29 - undefined. */
                                   {128, 255, 128},  /* 30 - undefined. */
                                   {128, 128, 255}}, /* 31 - undefined. */
      *colour_history_index = colour_history_start + 14;
const Colour *colour_history_end = colour_history_start + 32;

BITMAP *colour_selector_surface = 0;
void redrawColourSelector(BITMAP *bmp)
 {
  int u, v;
  Colour temp_colour;
  HSV temp_hsv;
  
  for(v = 0; v < 256; ++v)
   for(u = 0; u < 256; ++u)
    {
     temp_hsv.h = hsv.h;
     temp_hsv.s = v/255.0;
     temp_hsv.v = u/255.0;
     
     convertHSVToColour(temp_hsv, &temp_colour);
     _putpixel32(bmp, u, v, makecol32(temp_colour.r, temp_colour.g, temp_colour.b));
    }
  
  for(v = 0; v < 256; ++v)
   {
    temp_hsv.h = v*360.0/256.0;
    temp_hsv.s = 1.0;
    temp_hsv.v = 1.0;
    convertHSVToColour(temp_hsv, &temp_colour);
    hline(bmp, 256, v, 256+31, makecol32(temp_colour.r, temp_colour.g, temp_colour.b));
   }
  
  for(v = 0; v < 32; ++v)
   {
    rectfill(bmp, 256+32+(v/8)*32, (v&7)*32, 256+32+(v/8)*32+31, (v&7)*32+31, makecol32(colour_history_start[v].r*0.3, colour_history_start[v].g*0.3, colour_history_start[v].b*0.3));
    rectfill(bmp, 256+33+(v/8)*32, (v&7)*32+1, 256+32+(v/8)*32+30, (v&7)*32+30, makecol32(colour_history_start[v].r*0.7, colour_history_start[v].g*0.7, colour_history_start[v].b*0.7));
    rectfill(bmp, 256+34+(v/8)*32, (v&7)*32+2, 256+32+(v/8)*32+29, (v&7)*32+29, makecol32(colour_history_start[v].r, colour_history_start[v].g, colour_history_start[v].b));
   }
  
  drawHorizontalMarker(bmp, 256+16, hsv.h/360.0*256.0);
  drawCrossMarker(bmp, hsv.v*256.0, hsv.s*256.0);
  
  redrawCursor();
 }

BITMAP * paintOptions(BITMAP *bmp, int btn, int key, int x, int y)
 {
  int redraw = 0;
  
  if(!bmp)
   {
    if(!(bmp = colour_selector_surface = create_bitmap_ex(32, 256+32+32*4, 256)))
     return 0;
    else
     redraw = 1;
   }
  else colour_selector_surface = bmp;
  
  if(btn)
   {
    Colour temp;
    
    if(x < 256 && y < 256)
     {
      hsv.s = y/255.0;
      hsv.v = x/255.0;
      
      if(btn == 1)
       convertHSVToColour(hsv, &colour);
      else
       {
        convertHSVToColour(hsv, &temp);
        colour.r = colour.r*0.85 + temp.r*0.15;
        colour.g = colour.g*0.85 + temp.g*0.15;
        colour.b = colour.b*0.85 + temp.b*0.15;
        convertColourToHSV(colour, &hsv);
       }
      
      *colour_history_index = colour;
      if(++colour_history_index == colour_history_end)
       colour_history_index = colour_history_start;
     }
    else if(x >= 256 && x < 256+32)
     {
      hsv.h = y*360.0/256.0;
      
      if(btn == 1)
       convertHSVToColour(hsv, &colour);
      else
       {
        convertHSVToColour(hsv, &temp);
        colour.r = colour.r*0.85 + temp.r*0.15;
        colour.g = colour.g*0.85 + temp.g*0.15;
        colour.b = colour.b*0.85 + temp.b*0.15;
        convertColourToHSV(colour, &hsv);
       }
      
      *colour_history_index = colour;
      if(++colour_history_index == colour_history_end)
       colour_history_index = colour_history_start;
     }
    else if(btn == 1)
     convertColourToHSV(colour = colour_history_start[(x-(256+32))/32*8+y/32], &hsv);
    else
     {
      colour.r = colour.r*0.85 + colour_history_start[(x-(256+32))/32*8+y/32].r*0.15;
      colour.g = colour.g*0.85 + colour_history_start[(x-(256+32))/32*8+y/32].g*0.15;
      colour.b = colour.b*0.85 + colour_history_start[(x-(256+32))/32*8+y/32].b*0.15;
      convertColourToHSV(colour, &hsv);
     }
    
    redrawMenuIcons();
    redrawMenu();
    redraw = 1;
   }
  
  if(redraw)
   {
    redrawColourSelector(bmp);
    redrawCursor();
    return bmp;
   }
  
  return 0;
 }

typedef struct
 {
  const char *name, *imagename, *maskname;
  BITMAP * (*update)(BITMAP *bmp, int b, int k, int x, int y);
  int (*gencommand)(Command *command);
  BITMAP *image, *mask, *overlay;
  int masked_overlay; /* 0 = none, 1 = transparent, 2 = just masked. */
 } Menu;



Menu menu_start[] = {{"Pen",    "pen.bmp",    "penmask.bmp",      0,              &penGenerate,    0, 0, 0, 0},
                     {"Pencil", "pencil.bmp", "pencilmask.bmp",   &pencilOptions, &pencilGenerate, 0, 0, 0, 0},
                     {"Brush",  "brush.bmp",  "brushmask.bmp",    &brushOptions,  &brushGenerate,  0, 0, 0, 0},
                     {"Chalk",  "chalk.bmp",  "chalkmask.bmp",    &chalkOptions,  &chalkGenerate,  0, 0, 0, 0},
                     {"Smudge", "smudge.bmp", "smudgemask.bmp",   &smudgeOptions, &smudgeGenerate, 0, 0, 0, 0},
                     {"Stamp",  "stamp.bmp",  "stampmask.bmp",    &stampOptions,  &stampGenerate,  0, 0, 0, 0},
                     {"Paint",  "paint.bmp",  "paintmask.bmp",    &paintOptions,  0,               0, 0, 0, 0},
                     {"Chat",   "chat.bmp",   "chatmask.bmp",     &chatOptions,   0,               0, 0, 0, 1}},
     *menu_end = menu_start + sizeof menu_start / sizeof *menu_start;

const int CHAT_MENU_ID = 7; /* If the menus get rearranged, change this! */

int selected_menu = 0, last_drawable_menu = 0;

void menuCleanup(Menu *m)
 {
  assert(m >= menu_start && m < menu_end);
  
  if(m->image)
   {
    destroy_bitmap(m->image);
    m->image = 0;
   }
  
  if(m->mask)
   {
    destroy_bitmap(m->mask);
    m->mask = 0;
   }
  
  if(m->overlay)
   {
    destroy_bitmap(m->overlay);
    m->overlay = 0;
   }
 }

BITMAP *createPreviewBase(void)
 {
  uint32_t u, v;
  BITMAP *preview = create_bitmap_ex(32, 160, 120);
  
  if(!preview) return 0;
  
  drawPaper(preview);
  
  for(v = 0; v < 6; ++v)
   {
    Colour c = colour_history_start[RNG()%16];
    points = (RNG()%5)+10;
    
    for(u = 0; u < points; ++u)
     {
      x[u] = (RNG()%(preview->w-10))+5;
      y[u] = (RNG()%(preview->h-10))+5;
     }
    
    switch(RNG()%5)
     {
      case 0:
       drawPen(preview, c, x, y, points);
      break;
      
      case 1:
       drawPencil(preview, c, fRNG()*4.0+1.0, x, y, points);
      break;
      
      case 2:
       drawBrush(preview, c, fRNG()*4.0+1.0, fRNG()*1000.0, x, y, points);
      break;
      
      case 3:
       drawSmudge(preview, fRNG()*4.0+1.0, fRNG()*1000.0, x, y, points);
      break;
      
      case 4:
       drawChalk(preview, c, fRNG()*5.0+1.0, x, y, points);
      break;
     }
   }
  
  points = 0;
  return preview;
 }

void redrawMenuIcons(void)
 {
  int xoff = 0;
  Menu *m;
  
  clear_to_color(menu, makecol32(128, 172, 255));
  
  for(m = menu_start; m != menu_end; ++m, xoff+=42)
   {
    int x, y;
    
    if(!m->image && !(m->image = load_bitmap(m->imagename, 0)))
     {
      if(!(m->image = create_bitmap_ex(32, 42, 42)))
       continue; /* Can't draw it! */
      else
       { /* Draw an uber crappy default image. */
        clear_to_color(m->image, makecol32(64, 128, 172));
        
        #ifdef OLD_ALLEGRO
        textout_centre(m->image, font, m->name, 20, 20, makecol32(255, 255, 255));
        textout_centre(m->image, font, m->name, 22, 22, makecol32(0, 0, 0));
        textout_centre(m->image, font, m->name, 21, 21, makecol32(128, 128, 128));
        #else
        textout_centre_ex(m->image, font, m->name, 20, 20, makecol32(255, 255, 255), -1);
        textout_centre_ex(m->image, font, m->name, 22, 22, makecol32(0, 0, 0), -1);
        textout_centre_ex(m->image, font, m->name, 21, 21, makecol32(128, 128, 128), -1);
        #endif
       }
     }
    
    if(!m->mask && !(m->mask = load_bitmap(m->maskname, 0)))
     {
      if(!(m->mask = create_bitmap_ex(32, 42, 42)))
       continue; /* Can't draw it! */
      else
       { /* Draw an uber crappy default mask. */
        clear_to_color(m->mask, makecol32(0, 0, 0));
        
        #ifdef OLD_ALLEGRO
        textout_centre(m->mask, font, m->name, 21, 21, makecol32(255, 255, 255));
        #else
        textout_centre_ex(m->mask, font, m->name, 21, 21, makecol32(255, 255, 255), -1);
        #endif
       }
     }
    
    for(y = 0; y < 42; ++y) for(x = 0; x < 42; ++x)
     {
      int c = _getpixel32(m->mask, x, y);
      float beta = (getr32(c)+getb32(c)+getb32(c))/(255.0*3.0);
      c = _getpixel32(m->image, x, y);
      
      if(beta > 0.005)
       {
        float ib = 1.0-beta;
        int _r = (colour.r+getr32(c)-128.0)*beta+getr32(c)*ib,
            _g = (colour.g+getg32(c)-128.0)*beta+getg32(c)*ib,
            _b = (colour.b+getb32(c)-128.0)*beta+getb32(c)*ib;
        
        c = makecol32(_r>255?255:_r<0?0:_r, _g>255?255:_g<0?0:_g, _b>255?255:_b<0?0:_b);
       }
      
      _putpixel32(menu, x+xoff, y, c);
     }
   }
 }

float path_remainder = 0.0;
BITMAP *path_preview = 0;

void redrawPathPreview()
 {
  static Region last_region = {0, 0, 0, 0};
  
  if(points > 1)
   {
    blit(path_preview, buffer, last_region.x1-(screen_w-buffer_w)/2, last_region.y1-(screen_h-buffer_h)/2, last_region.x1, last_region.y1, last_region.x2-last_region.x1, last_region.y2-last_region.y1);
    DRAdd(&dirt, &last_region);
    drawPreviewLine(buffer, colour, x, y, points, points == 2 ? 0 : points-3, points, path_remainder, &last_region);
    DRAdd(&dirt, &last_region);
   }
 }

void redrawMenu(void)
 {
  Region menu_region;
  
  redrawAllUsers();
  blit(menu, buffer, 0, 0, 0, screen_h-42, screen_w, 42);
  rect(buffer, selected_menu*42+1, screen_h-41, selected_menu*42+41, screen_h-1, makecol32(0, 0, 0));
  rect(buffer, selected_menu*42, screen_h-42, selected_menu*42+40, screen_h-2, makecol32(255, 255, 255));
  
  menu_region.x1 = 0;
  menu_region.y1 = screen_h-43;
  menu_region.x2 = screen_w;
  menu_region.y2 = screen_h;
  
  DRAdd(&dirt, &menu_region);
 }

void redrawMenuOverlay(void)
 {
  if(menu_start[selected_menu].overlay)
   {
    Region overlay;
    
    int overlay_x = selected_menu*42+21-menu_start[selected_menu].overlay->w/2;
    if(overlay_x < 0) overlay_x = 0;
    else if(overlay_x > screen_w-menu_start[selected_menu].overlay->h) overlay_x = screen_w-menu_start[selected_menu].overlay->h;
    
    
    overlay.x1 = overlay_x;
    overlay.y1 = screen_h-42-menu_start[selected_menu].overlay->h;
    overlay.x2 = overlay_x+menu_start[selected_menu].overlay->w;
    overlay.y2 = screen_h-42;
    
    DRAdd(&dirt, &overlay);
    
    switch(menu_start[selected_menu].masked_overlay)
     {
      case 1:
      case 2:
       rectfill(buffer, overlay_x, screen_h-42-menu_start[selected_menu].overlay->h, overlay_x+menu_start[selected_menu].overlay->w-1, screen_h-43, 0);
       
       blit(local_image, buffer,
            overlay_x-(screen_w-buffer_w)/2, (screen_h-42-menu_start[selected_menu].overlay->h)-(screen_h-buffer_h)/2,
            overlay_x, screen_h-42-menu_start[selected_menu].overlay->h, menu_start[selected_menu].overlay->w, menu_start[selected_menu].overlay->h);
       
       if(menu_start[selected_menu].masked_overlay == 1)
        {
         int x, y = screen_h-42-menu_start[selected_menu].overlay->h, 
             ex = overlay_x+menu_start[selected_menu].overlay->w, ey = screen_h-42;
         
         for(; y < ey; ++y) for(x = overlay_x; x < ex; ++x)
          {
           int c = _getpixel32(buffer, x, y),
           v = (getr32(c)+getg32(c)+getb32(c))*2;
           
           _putpixel32(buffer, x, y, makecol32((v+getr32(c))/14, (v+getg32(c))/12, (v+getb32(c))/10));
          }
        }
       
       masked_blit(menu_start[selected_menu].overlay, buffer, 0, 0, overlay_x, screen_h-42-menu_start[selected_menu].overlay->h, menu_start[selected_menu].overlay->w, menu_start[selected_menu].overlay->h);
      break;
      
      default:
       blit(menu_start[selected_menu].overlay, buffer, 0, 0, overlay_x, screen_h-42-menu_start[selected_menu].overlay->h, menu_start[selected_menu].overlay->w, menu_start[selected_menu].overlay->h);
     }
    (menu_start[selected_menu].masked_overlay?masked_blit:blit)(menu_start[selected_menu].overlay, buffer, 0, 0, overlay_x, screen_h-42-menu_start[selected_menu].overlay->h, menu_start[selected_menu].overlay->w, menu_start[selected_menu].overlay->h);
   }
 }

void undrawMenuIcon(int icon)
 {
  Region region;
  
  blit(menu, buffer, icon*42, 0, icon*42, screen_h-42, 42, 42);
  
  region.x1 = icon*42;
  region.y1 = screen_h-42;
  region.x2 = icon*42+42;
  region.y2 = screen_h;
  
  DRAdd(&dirt, &region);
 }

void redrawMenuIcon(int icon)
 {
  Region region;
  
  blit(menu, buffer, icon*42, 0, icon*42, screen_w-42, 42, 42);
  rect(buffer, icon*42+1, screen_h-41, icon*42+41, screen_h-1, makecol32(0, 0, 0));
  rect(buffer, icon*42, screen_h-42, icon*42+40, screen_h-2, makecol32(255, 255, 255));
  
  region.x1 = icon*42;
  region.y1 = screen_h-42;
  region.x2 = icon*42+42;
  region.y2 = screen_h;
  
  DRAdd(&dirt, &region);
 }

void redrawCanvas(BITMAP *bmp)
 {
  Region region;
  
  assert(bmp);
  assert(bmp->w == buffer_w && bmp->h == buffer_h);
  
  blit(bmp, buffer, 0, 0, (screen_w-buffer_w)/2, (screen_h-buffer_h)/2, buffer_w, buffer_h);
  
  if(screen_h > buffer_h)
   {
    rectfill(buffer, 0, 0, screen_w, (screen_h-buffer_h)/2-1, makecol32(0, 0, 0));
    rectfill(buffer, 0, screen_h-(screen_h-buffer_h)/2, screen_w, screen_h-1, makecol32(0, 0, 0));
   }
  
  if(screen_w > buffer_w)
   {
    rectfill(buffer, 0, (screen_h-buffer_h)/2, (screen_w-buffer_w)/2-1, screen_h-(screen_h-buffer_h)/2-1, makecol32(0, 0, 0));
    rectfill(buffer, screen_w-(screen_w-buffer_w)/2, (screen_h-buffer_h)/2, screen_w-1, screen_h-(screen_h-buffer_h)/2-1, makecol32(0, 0, 0));
   }
  
  drew_users = 0;
  
  region.x1 = 0;
  region.y1 = 0;
  region.x2 = screen_w;
  region.y2 = screen_h;
  
  DRAdd(&dirt, &region);
 }

void displayMessage(const char *string, ...)
 {
  char buf[512];
  va_list args;
  va_start(args, string);
  vsnprintf(buf, sizeof buf, string, args);
  va_end(args);
  
  chatAddMessage(0, buf);
 }

#ifdef OLD_ALLEGRO
volatile int shutdown_flag = 0;

void shutdownCallback(void)
 {
  shutdown_flag = 1;
 }
#endif

int main(int argc, char *argv[])
 {
  int menu_click = 0, sync_marks = 0;
  float distance = 0;
  Command command;
  Region marks;
  Menu *m;
  
  DRInit(&dirt);
  
  logStart();
  
  #ifdef OLD_ALLEGRO
  text_mode(-1);
  #endif
  
  can_connect = networkInit();
  
  allegro_init();
  install_keyboard();
  install_mouse();
  set_color_depth(32);
  
  if(!configLoad("client.cfg"))
   {
    if(set_gfx_mode(GFX_AUTODETECT_WINDOWED, 320, 240, 0, 0))
     return EXIT_FAILURE;
    else if(!configCreate("client.cfg"))
     return EXIT_FAILURE;
   }
  
  if(set_gfx_mode(fullscreen ? GFX_AUTODETECT : GFX_AUTODETECT_WINDOWED, screen_w, screen_h, 0, 0))
   {
    allegro_message("Error initiating graphics. Try deleting client.cfg.");
    return EXIT_FAILURE;
   }
  
  marks.x1 = screen_w/2-30;
  marks.y1 = 1;
  marks.x2 = screen_w/2+30;
  marks.y2 = 5;
  
  #ifdef USE_ALFONT
  if(alfont_init() == ALFONT_OK)
   {
    if(!(alfont = alfont_load_font("DejaVuSans-Bold.ttf")))
     alfont_exit();
    else
     alfont_set_font_size(alfont, 18);
   }
  #endif
  
  #ifdef OLD_ALLEGRO
  set_window_close_hook(&shutdownCallback);
  #endif
  
  preview = createPreviewBase();
  buffer = create_bitmap_ex(32, screen_w, screen_h);
  scratch_buffer = create_bitmap_ex(32, screen_w, screen_h);
  remote_image = create_bitmap_ex(32, buffer_w, buffer_h);
  local_image = create_bitmap_ex(32, buffer_w, buffer_h);
  path_preview = create_bitmap_ex(32, buffer_w, buffer_h);
  menu = create_bitmap_ex(32, screen_w, 42);
  cursor = create_bitmap_ex(32, 6, 6);
  
  if(!preview || !buffer || !scratch_buffer || !path_preview || !local_image || !remote_image || !menu || !cursor)
   {
    set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
    allegro_message("Error creating image buffers.");
    return EXIT_FAILURE;
   }
  
  drawPaper(remote_image);
  
  sRNG(time(0));
  
  blit(remote_image, local_image, 0, 0, 0, 0, buffer_w, buffer_h);
  clear(buffer);
  redrawCursor();
  redrawMenuIcons();
  
  if(!commandQueueInit(&pending))
   {
    set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
    allegro_message("Error initiating command queue.");
   }
  
  if(can_connect)
   {
    displayMessage("Attempting to connect to %s:%d. . .", server, port);
    
    if(!networkConnectToServer(&connection, server, port))
     connected = 0;
    else
     {
      commandInit(&command, Login);
      
      if(!commandPutU32(&command, VERSION) || !commandPutString(&command, username))
       {
        connected = 0;
        commandUninit(&command);
        networkKillConnection(&connection);
       }
      else if(!commandNetworkSend(&command, &connection))
       {
        connected = 0;
        commandUninit(&command);
        networkKillConnection(&connection);
       }
      else if(!commandQueuePushBack(&pending, &command))
       {
        connected = 0;
        commandUninit(&command);
        networkKillConnection(&connection);
       }
      else connected = 1;
     }
    
    if(!connected)
     displayMessage("Unable to connect to %s:%d.", server, port);
   }
  
  redrawCanvas(local_image);
  
  #ifdef OLD_ALLEGRO
  while(!shutdown_flag)
  #else
  while(1)
  #endif
   {
    int need_refresh = 0;
    int button = 0;
    int key_code = keypressed() ? readkey() : 0;
    static int old_canvas_x = 0, old_canvas_y = 0;
    int canvas_x = mouse_x-(screen_w-buffer_w)/2, canvas_y = mouse_y-(screen_h-buffer_h)/2;
    
    poll_mouse();
    poll_keyboard();
    
    if((key_code >> 8) == KEY_TAB ||
       ((!inmenu ||
        (inmenu && selected_menu != CHAT_MENU_ID)) &&
        ((key_code >> 8) == KEY_SLASH || (key_code >> 8) == KEY_TILDE || (key_code >> 8) ==  KEY_SLASH_PAD)) || (inmenu && (key_code >> 8) == KEY_ESC))
     {
      if(inmenu && ((key_code >> 8) == KEY_TAB || (key_code >> 8) == KEY_ESC))
       {
        chatRedrawOverlay(buffer, local_image, 1);
        inmenu = 0;
       }
      else
       {
        if(points)
         {
          points = 0;
          chatRedrawOverlay(buffer, local_image, 1);
         }
        
        if((key_code >> 8) == KEY_SLASH || (key_code >> 8) == KEY_TILDE || (key_code >> 8) == KEY_SLASH_PAD)
         selected_menu = CHAT_MENU_ID;
        
        redrawMenu();
        redrawMenuOverlay();
        inmenu = 1;
       }
      
      key_code = 0;
     }
    
    if((key_code >> 8) == KEY_ESC)
     break;
    
    if((mouse_b&1) || (mouse_b&2))
     {
      if(!menu_click)
       {
        button = (mouse_b&2)?2:1;
        menu_click = 1;
       }
     }
    else menu_click = 0;
    
    if(inmenu)
     {
      BITMAP *overlay = 0;
      
      if(menu_start[selected_menu].update)
       {
        if(menu_start[selected_menu].overlay)
         {
          int overlay_x = selected_menu*42+21-menu_start[selected_menu].overlay->w/2, x, y;
          if(overlay_x < 0) overlay_x = 0;
          else if(overlay_x > screen_w-menu_start[selected_menu].overlay->h) overlay_x = screen_w-menu_start[selected_menu].overlay->h;
          
          x = mouse_x - overlay_x;
          y = mouse_y - (screen_h-42-menu_start[selected_menu].overlay->h);
          
          if(x >= 0 && y >= 0 && x < menu_start[selected_menu].overlay->w && y < menu_start[selected_menu].overlay->h)
           {
            overlay = (*menu_start[selected_menu].update)(menu_start[selected_menu].overlay, button, key_code, x, y);
            button = 0;
           }
          else
           overlay = (*menu_start[selected_menu].update)(menu_start[selected_menu].overlay, 0, key_code, 0, 0);
         }
        else overlay = (*menu_start[selected_menu].update)(0, 0, key_code, 0, 0);
        
        if(overlay)
         {
          menu_start[selected_menu].overlay = overlay;
          redrawMenuOverlay();
         }
       }
      
      if(button)
       {
        if(mouse_y > screen_h-42)
         {
          if(menu_start + mouse_x/42 < menu_end && selected_menu != mouse_x/42)
           {
            selected_menu = mouse_x/42;
            
            if(menu_start[selected_menu].gencommand)
             last_drawable_menu = selected_menu;
            
            chatRedrawOverlay(buffer, local_image, 1);
            redrawMenu();
            redrawMenuOverlay();
           }
          
          button = 0;
         }
        else
         {
          chatRedrawOverlay(buffer, local_image, 1);
          inmenu = 0;
         }
       }
      
      key_code = 0;
     }
    
    if(can_connect && connected)
     {
      if(!networkUpdateConnection(&connection))
       {
        networkKillConnection(&connection);
        connected = 0;
        points = 0;
        displayMessage("You have been disconnected from %s:%d", server, port);
       }
      else
       {
        time_t start = time(0);
        
        while(time(0)-start < 1 && commandInitFromNetwork(&command, &connection))
         {
          switch(command.type)
           {
            case RepeatCommand:
             
             logMessage(Debug, "Our last command has been accepted by the server.");
             commandUninit(&command);
             if(commandQueuePopFront(&pending, &command))
              {
               if(need_refresh ||
                  command.type == Chat || /* We have to draw chat messages now because they're not buffered; they don't appear until the server accepts them. */
                  commandQueueElements(&pending))
                commandDraw(remote_image, &command);
               else
                blit(local_image, remote_image, 0, 0, 0, 0, buffer_w, buffer_h);
               
               commandUninit(&command);
              }
             else logMessage(Debug, ". . . Wait a minute, we didn't send that command. What the heck!");
            
            break;
            
            case Login:
             {
              uint8_t id, user_count = 0, last_id = 0;
              
              userForgetAll();
              userid = commandGetU8(&command);
              commandGetString(&command, username, sizeof username);
              userSetName(userid, username);
              
              while((id = commandGetU8(&command)))
               {
                char name[16];
                commandGetString(&command, name, sizeof name);
                userSetName(id, name);
                ++user_count;
                last_id = id;
               }
              
              if(user_count == 0)
               displayMessage("Logged in as %s.", username);
              else if(user_count == 1)
               displayMessage("Logged in as %s, %s is already here.", username, userGetName(last_id));
              else
               displayMessage("Logged in as %s, %d users are already here.", username, user_count);
              
              commandUninit(&command);
              drawPaper(remote_image);
              need_refresh = 1;
             }
            
            break;
            
            case Join:
             {
              uint8_t id = commandGetU8(&command);
              char name[16];
              char message[128];
              commandGetString(&command, name, sizeof name);
              userSetName(id, name);
              logMessage(Debug, "New user %d (%s) joined.", id, name);
              
              snprintf(message, sizeof message, "%s has joined.", name);
              chatAddMessage(0, message);
              
              commandUninit(&command);
             }
            break;
            
            case Leave:
             {
              uint8_t id = commandGetU8(&command);
              char message[128];
              
              logMessage(Debug, "User %d (%s) left.", id, userGetName(id));
              
              snprintf(message, sizeof message, "%s has left.", userValid(id) ? userGetName(id) : "???");
              chatAddMessage(0, message);
              
              userForget(id);
              
              /* TODO: Set all draw commands owned by this user to 0 when we start storing them. */
              
              commandUninit(&command);
             }
            break;
            
            case Restart:
             {
              chatAddMessage(0, "Image restarted.");
              drawPaper(remote_image);
              need_refresh = 1;
              commandUninit(&command);
             }
            break;
            
            default:
             commandDraw(remote_image, &command);
             commandUninit(&command);
             need_refresh = 1;
           }
         }
        
        if(need_refresh)
         {
          uint16_t c;
          
          logMessage(Debug, "Need to redraw buffer for newly recieved data.");
          
          blit(remote_image, local_image, 0, 0, 0, 0, buffer_w, buffer_h);
          
          for(c = 0; c < commandQueueElements(&pending); ++c)
           if(commandQueueElement(&pending, c)->type != Chat) /* Drawing chat commands would make them appear multiple times. That would be bad. */
            commandDraw(local_image, commandQueueElement(&pending, c));
          
          if(points)
           {
            blit(local_image, path_preview, 0, 0, 0, 0, buffer_w, buffer_h);
            
            if(points > 3)
             path_remainder = drawPreviewLine(path_preview, colour, x, y, points, 0, points-3, 0.0, 0);
            
            chatRedrawOverlay(buffer, path_preview, 1);
            redrawPathPreview();
           }
          else chatRedrawOverlay(buffer, local_image, 1);
          
          if(inmenu)
           {
            redrawMenu();
            redrawMenuOverlay();
           }
         }
       }
     }
    
    if(button == 1)
     {
      points = 2;
      distance = 0.0;
      path_remainder = 0.0;
      x[0] = x[1] = canvas_x;
      y[0] = y[1] = canvas_y;
      blit(local_image, path_preview, 0, 0, 0, 0, buffer_w, buffer_h);
     }
    else if(points)
     {
      int canvas_dx, canvas_dy;
      
      if(key[KEY_LCONTROL] || key[KEY_RCONTROL])
       {
        canvas_x = (canvas_x+6)/12*12;
        canvas_y = (canvas_y+6)/12*12;
       }
      
      canvas_dx = canvas_x - old_canvas_x;
      canvas_dy = canvas_y - old_canvas_y;
      
      distance += sqrt(canvas_dx*canvas_dx+canvas_dy*canvas_dy);
      
      if(!key[KEY_LSHIFT] && !key[KEY_RSHIFT] && distance >= 3.0 && 
          sqrt((canvas_x-x[points-2])*(canvas_x-x[points-2])+(canvas_y-y[points-2])*(canvas_y-y[points-2])) > 1.8 &&
         (points < 3 || distance > 100.0 || angleDiff(canvas_x-x[points-2], canvas_y-y[points-2], x[points-2]-x[points-3], y[points-2]-y[points-3])*distance > 5.0))
       {
        distance = 0;
        
        x[points-1] = x[points] = canvas_x;
        y[points-1] = y[points] = canvas_y;
        
        if(++points >= PATH_MAX_POINTS)
         --points;
        else if(points > 3)
         path_remainder = drawPreviewLine(path_preview, colour, x, y, points, points-4, points-3, path_remainder, 0);
       }
      else
       {
        x[points-1] = canvas_x;
        y[points-1] = canvas_y;
       }
      
      if(mouse_b & 2)
       {
        points = 0;
        chatRedrawOverlay(buffer, local_image, 1);
       }
      else if(mouse_b & 1)
       { /* Preview path. */
        if(canvas_x != old_canvas_x || canvas_y != old_canvas_y)
         redrawPathPreview();
       }
      else if((points == 2 && (x[0] != x[1] || y[0] != y[1])) || points > 2)
       { /* Officially draw the path. */
        if(!menu_start[last_drawable_menu].gencommand)
         goto done_drawing;
        
        if(!(*menu_start[last_drawable_menu].gencommand)(&command))
         {
          displayMessage("Error initiating %s command.", menu_start[last_drawable_menu].name);
          goto done_drawing;
         }
        
        commandDraw(local_image, &command);
        
        if(can_connect && connected)
         {
          if(!commandNetworkSend(&command, &connection))
           {
            commandUninit(&command);
            networkKillConnection(&connection);
            connected = 0;
            displayMessage("Disconnected from %s:%d", server, port);
           }
          else if(!commandQueuePushBack(&pending, &command))
           {
            commandUninit(&command);
            networkKillConnection(&connection);
            connected = 0;
            displayMessage("Out of memory; forced to disconnect from %s:%d", server, port);
           }
         }
        else commandUninit(&command);
        
        done_drawing:
        
        chatRedrawOverlay(buffer, local_image, 1);
        points = 0;
       }
      else points = 0;
     }
    else if(button == 2)
     {
      getSampleColour(local_image, mouse_x-(screen_w-buffer_w)/2, mouse_y-(screen_h-buffer_h)/2, &colour);
      convertColourToHSV(colour, &hsv);
      redrawCursor();
      redrawMenuIcons();
      
      if(colour_selector_surface)
       redrawColourSelector(colour_selector_surface);
     }
    
    old_canvas_x = canvas_x;
    old_canvas_y = canvas_y;
    
    if(can_connect && connected && commandQueueElements(&pending))
     {
      if(++sync_marks == 50)
       sync_marks = 1;
      
      if(need_refresh || (sync_marks-1)/5 != (sync_marks-2)/5)
       {
        int u;
        
        for(u = 0; u < 60; ++u)
         vline(buffer, screen_w/2-30+u, 1, 4, (((u+(sync_marks-1)/5)/5)&1)?makecol32(0,0,0):makecol32(255,0,0));
        
        DRAdd(&dirt, &marks);
       }
     }
    else if(sync_marks)
     {
      if(1-(screen_h-buffer_h)/2 < 0)
       rectfill(buffer, screen_w/2-30, 1, screen_w/2+29, 4, makecol32(0, 0, 0));
      else
       blit(local_image, buffer, screen_w/2-30-(screen_w-buffer_w)/2, 1-(screen_h-buffer_h)/2, screen_w/2-30, 1, 60, 4);
      
      DRAdd(&dirt, &marks);
      sync_marks = 0;
     }
    
    if(chatRedrawOverlay(buffer, points ? path_preview : local_image, 0))
     {
      if(inmenu)
       {
        redrawMenu();
        redrawMenuOverlay();
       }
      else if(points)
       redrawPathPreview();
     }
    
    if((key[KEY_ALT] || key[KEY_ALTGR]) && !inmenu)
     {
      if(!drew_users)
       {
        redrawAllUsers();
        drew_users = 1;
       }
     }
    else if(drew_users && !inmenu)
     {
      chatRedrawOverlay(buffer, points?path_preview:local_image, 1);
      drew_users = 0;
     }
    
    if(!DRIsClean(&dirt))
     {
      Region region;
      
      region.x1 = 0;
      region.y1 = 0;
      region.x2 = buffer->w;
      region.y2 = buffer->h;
      
      DRClip(&dirt, &region);
      DROptimize(&dirt);
      
      if(DRIsSaturated(&dirt)) /* There is more dirt than the structure could hold, redraw the entire screen. */
       {
        scare_mouse();
        blit(buffer, screen, 0, 0, 0, 0, buffer->w, buffer->h);
        unscare_mouse();
       }
      else 
       {
        const Region *r=dirt.start;
        
        for(; r != dirt.end; ++r)
         {
          scare_mouse_area(r->x1, r->y1, r->x2-r->x1, r->y2-r->y1);
          blit(buffer, screen, r->x1, r->y1, r->x1, r->y1, r->x2-r->x1, r->y2-r->y1);
          unscare_mouse();
         }
       }
      
      DRMakeClean(&dirt);
     }
    
    #ifdef OLD_ALLEGRO
    yield_timeslice();
    #else
    rest(0);
    #endif
   }
  
  for(m = menu_start; m != menu_end; ++m)
   menuCleanup(m);
  
  destroy_bitmap(local_image);
  destroy_bitmap(remote_image);
  destroy_bitmap(buffer);
  destroy_bitmap(path_preview);
  destroy_bitmap(menu);
  destroy_bitmap(cursor);
  
  #ifdef USE_ALFONT
  if(alfont)
   {
    alfont_destroy_font(alfont);
    alfont = 0;
    alfont_exit();
   }
  #endif
  
  allegro_exit();
  
  if(can_connect)
   {
    networkUninit();
    
    if(connected)
     networkKillConnection(&connection);
   }
  
  commandQueueUninit(&pending);
  logStop();
  
  return EXIT_SUCCESS;
 } END_OF_MAIN()
