#ifndef DRAW_H
#define DRAW_H

#include <allegro.h>
#include <stdint.h>
#include "region.h"

typedef struct {float r, g, b;} Colour;
typedef struct {float h, s, v;} HSV;

void convertColourToHSV(Colour, HSV *);
void convertHSVToColour(HSV, Colour *);

void getSampleColour(BITMAP *surface, int16_t x, int16_t y, Colour *c);

void drawPaper(BITMAP *surface);
float drawPreviewLine(BITMAP *surface, Colour c, const int16_t *x, const int16_t *y, uint32_t points, uint32_t first, uint32_t last, float remainder, Region *r);

void drawPen(BITMAP *surface, Colour c, const int16_t *x, const int16_t *y, uint32_t points);
void drawPencil(BITMAP *surface, Colour c, float radius, const int16_t *x, const int16_t *y, uint32_t points);
void drawBrush(BITMAP *surface, Colour c, float radius, float rate, const int16_t *x, const int16_t *y, uint32_t points);
void drawSmudge(BITMAP *surface, float radius, float rate, const int16_t *x, const int16_t *y, uint32_t points);
void drawChalk(BITMAP *surface, Colour c, float radius, const int16_t *x, const int16_t *y, uint32_t points);
void drawStamp(BITMAP *surface, Colour c, float radius, float rate, const int16_t *x, const int16_t *y, uint32_t points);

extern const float brush_min_radius, brush_max_radius, brush_min_rate, brush_max_rate;
extern const float pencil_min_radius, pencil_max_radius;
extern const float smudge_min_radius, smudge_max_radius, smudge_min_rate, smudge_max_rate;
extern const float chalk_min_radius, chalk_max_radius;
extern const float stamp_min_radius, stamp_max_radius, stamp_min_rate, stamp_max_rate;

#endif
