#include "common.h"
#include "guibits.h"

// Common bits and pieces used only here.

static BITMAP *tl = 0, *t = 0,  *tr = 0,
              *l = 0,  *bg = 0, *r = 0,
              *bl = 0, *b = 0,  *br = 0,
              *resize = 0;

static int init_count = 0;

static void unloadGUIBits(void)
 {
  assert(init_count > 0);
  
  if(--init_count == 0)
   {
    #define KILLBMP(A) if(A) {destroy_bitmap(A); A = 0;}
    KILLBMP(tl)
    KILLBMP(t)
    KILLBMP(tr)
    KILLBMP(l)
    KILLBMP(bg)
    KILLBMP(r)
    KILLBMP(bl)
    KILLBMP(b)
    KILLBMP(br)
    KILLBMP(resize)
    #undef KILLBMP
   }
 }

static int loadGUIBits(void)
 {
  if(++init_count == 1)
   {
    #define MAKEBMP(A) if(!(A = load_bitmap(#A".bmp", 0))) {unloadGUIBits(); return 0;}
    MAKEBMP(tl)
    MAKEBMP(t)
    MAKEBMP(tr)
    MAKEBMP(l)
    MAKEBMP(bg)
    MAKEBMP(r)
    MAKEBMP(bl)
    MAKEBMP(b)
    MAKEBMP(br)
    MAKEBMP(resize)
    #undef MAKEBMP
   }
  
  return 1;
 }

static void blitHTile(BITMAP *src, BITMAP *dest, int x, int y, int w)
 {
  assert(src);
  assert(dest);
  
  for(; x < w ; x += src->w)
   blit(src, dest, 0, 0, x, y, w - x, src->h);
 }

static void blitVTile(BITMAP *src, BITMAP *dest, int x, int y, int h)
 {
  assert(src);
  assert(dest);
  
  for(; y < h ; y += src->h)
   blit(src, dest, 0, 0, x, y, src->w, h - y);
 }

static void blitTile(BITMAP *src, BITMAP *dest, int _x, int y, int w, int h)
 {
  assert(src);
  assert(dest);
  
  int x;
  
  for(; y < h ; y += src->h)
   for(x = _x; x < w; x += src->w)
    blit(src, dest, 0, 0, x, y, w - x, h - y);
 }

// End common bits and pieces used only here.

void redrawBaseWindow(Window *w)
 {
  assert(w);
  
  // Draw the corners.
  blit(tl, w->buffer, 0, 0, 0, 0, tl->w, tl->h);
  blit(tr, w->buffer, 0, 0, w->buffer->w-tr->w, 0, tr->w, tr->h);
  blit(bl, w->buffer, 0, 0, 0, w->buffer->h-tr->h, bl->w, bl->h);
  blit(br, w->buffer, 0, 0, w->buffer->w-tr->w, w->buffer->h-tr->h, br->w, br->h);
  
  // Draw the edges.
  blitHTile(t, w->buffer, tl->w, 0, w->buffer->w-tr->w);
  blitHTile(b, w->buffer, bl->w, w->buffer->h-b->h, w->buffer->w-br->w);
  blitVTile(l, w->buffer, 0, tl->h, w->buffer->h-tl->h);
  blitVTile(r, w->buffer, w->buffer->w-r->w, tr->h, w->buffer->h-tr->h);
  
  // Draw the surface.
  blitTile(bg, w->buffer, tl->w, tl->h, w->buffer->w-br->w, w->buffer->h-br->h);
 }

void destroyBaseWindow(Window *w)
 {
  // Note: ResizeWindow also uses this function. If we do anything else here later, this may need to change.
  assert(w);
  unloadGUIBits();
 }

int BaseWindow(Window *w, void *title)
 {
  assert(w);
  
  if(title) { /* Todo. . . */ }
  
  w->redraw = &redrawBaseWindow;
  w->destroy = &destroyBaseWindow;
  w->attr |= attr_resize | attr_masked;
  
  if(!loadGUIBits())
   return 0;
  
  Window *resizer = GUICreateWindow(&ResizeWindow, 0, w->buffer->w-resize->w, w->buffer->h-resize->h, resize->w, resize->h);
  
  if(!resizer)
   {
    unloadGUIBits();
    return 0;
   }
  
  if(!GUIAddChild(w, resizer))
   {
    GUIDestroyWindow(resizer);
    unloadGUIBits();
    return 0;
   }
  
  return 1;
 }

void redrawResizeWindow(Window *w)
 {
  assert(w);
  
  blit(resize, w->buffer, 0, 0, w->buffer->w-resize->w, w->buffer->h-resize->h, resize->w, resize->h);
 }

int mouseResizeWindow(Window *w, int x, int y, int xoff, int yoff, int b, int b_off)
 {
  if((b & 1) && w->parent && (xoff || yoff))
   {
    int width = w->parent->buffer->w + xoff,
        height = w->parent->buffer->h + yoff;
    
    if(width < resize->w) width = resize->w;
    if(height < resize->h) height = resize->h;
    
    logMessage(Debug, "Resized to %dx%d.", width, height);
    
    GUIResizeWindow(w->parent, width, height);
    GUIPositionWindow(w, w->parent->buffer->w-resize->w, w->parent->buffer->h-resize->h);
    return 1;
   }
  
  return 0;
 }

int ResizeWindow(Window *w, void *unused)
 {
  w->redraw = &redrawResizeWindow;
  w->destroy = &destroyBaseWindow;
  w->mouse = &mouseResizeWindow;
  w->attr |= attr_masked | attr_select;
  
  if(!loadGUIBits())
   return 0;
  
  return 1;
 }
